import ast
import re
from datetime import datetime
from typing import Any, Callable, Dict, List, Union

from ..base import Ch, Downloader, Manga, Search, soup

template = "generic"

def chapter(url) -> Ch:
    c = soup("/".join(url.split("/")[:-1])).select_one(f'a[href="{url}"]').find_parent()
    a = c.select_one("a")
    ch = a["href"].split("-")[-1]
    ch = ast.literal_eval(ch) if ch else ch
    return Ch(
        url             = url,
        ch              = ch,
        vol             = None,
        title           = a.text,
        views           = int(c.select_one("span.chapter-view.text-nowrap").text.replace(",", "")),
        uploaded_at     = str(datetime.strptime(c.select_one("span.chapter-time.text-nowrap")["title"], '%b %d,%Y %H:%M')),
        imgs            = [i["src"] for i in soup(url).select(".container-chapter-reader img")]
    )

def manga(url: str, chs: int=0) -> Manga:
    ms = soup(url)
    ti = [i["class"][0] for i in ms.select(".table-label i")]
    si = [i["class"][0] for i in ms.select(".stre-label i")]
    def info(key: str, l: Callable[[str], Any]=None, d: Any=None) -> Any:
        key = f"info-{key}"
        op = d
        if key in ti:
            op = ms.select(f'.variations-tableInfo tbody tr:nth-child({ti.index(key) + 1}) td')[1]
        elif key in si:
            op = ms.select(f'p:nth-child({si.index(key) + 1}) .stre-value')[0]
        return l(op) if l else d
    rc = ms.select("li.a-h")
    chap_dict = {}
    if chs:
        n = 1
        for c in rc:
            a = c.select_one("a")
            ch = a["href"].split("-")[-1]
            if ch:
                k = ast.literal_eval(ch)
            else:
                k = -n
                n += 1
            cch = a["href"]
            if chs == 2:
                cch = chapter(cch)
            chap_dict[k] = cch
    return Manga(
        url             = url,
        title           = ms.select_one(".story-info-right h1").text,
        author          = info("author", lambda x: [i.text for i in x.select("a") if i], []),
        covers          = [ms.select_one(".info-image > img")["src"]],
        alt_titles      = info("alternative", lambda x: [x.text], []),
        status          = info("status", lambda x: {k: v for v, k in enumerate(["ongoing", "completed", "hiatus", "cancelled"])}.get(x.text.lower()), -1),
        genres          = info("genres", lambda x: {i.text: i["href"] for i in x.select("a") if i}, None),
        updated_at      = info("time", lambda x: str(datetime.strptime(x.text[:-3], '%b %d,%Y - %H:%M')), ""),
        created_at      = str(datetime.strptime(rc[-1].select_one("span.chapter-time.text-nowrap")["title"], '%b %d,%Y %H:%M')),
        views           = info("view", lambda x: int(x.text.replace(',', '')), 0),
        description     = ms.select_one(".panel-story-info-description").text,
        chapters        = chap_dict,
    )

def dl_search(title: str, **kwargs: Dict[str, Any]) -> Dict[str, str]:
    title = re.sub(r'[^A-Za-z0-9 ]+', '', title).replace(" ", "_")
    sr = {}
    ms = soup(f"https://manganato.com/search/story/{title}")
    gp = ms.select_one(".group-page")
    if gp:
        for p in range(len(gp.select("a")) - 2):
            for r in soup(f"https://manganato.com/search/story/hello?page={p+1}").select(".a-h.text-nowrap.item-title"):
                sr[r["title"]] = r["href"]
    else:
        for r in ms.select(".a-h.text-nowrap.item-title"):
            sr[r["title"]] = r["href"]
    return sr

def search(s: Search) -> List[Manga]:
    return [manga(i) for i in dl_search(**s.__dict__).values()]

def cli_search(title: str, **kwargs: Dict[str, Any]):
    return search(Search(title, **kwargs))

def ch_fn(url: str) -> List[str]:
    return [i["src"] for i in soup(url).select(".container-chapter-reader img")]

def chdls(url: str) -> List[Dict[Union[float, int, None], str]]:
    op = []
    for c in soup(url).select("li.a-h"):
        a = c.select_one("a")
        chdls.append({a["href"].split("-")[-1]: a["href"]})
    return op

def dl(url: str, **kwargs: Dict[str, Any]):
    Downloader(ch_fn, headers={"Referer": "https://readmanganato.com/"}, **kwargs).dl_chdls(chdls(url), soup(url).select_one(".story-info-right h1").text)

def cli_dl(title: str, **kwargs: Dict[str, Any]):
    Downloader(ch_fn, headers={"Referer": "https://readmanganato.com/"}, **kwargs).cli(cli_search, chdls, title)



from functools import partial
from importlib import import_module
from types import ModuleType
from typing import Any, Callable, Dict, List, Union
from urllib.parse import quote_plus

from bs4 import BeautifulSoup

from ....utils.utils import dt, sanitize_text
from ...base import Ch, Downloader, Manga, soup


class rch_fn:
    def setsu(url: str, base_url: str, manga_id_fn: Callable[[BeautifulSoup], Union[int, float]]):
        data = {"action": "manga_get_chapters", "manga": str(manga_id_fn(soup(url)))}
        return soup(f"{base_url}/wp-admin/admin-ajax.php", method="post", data=data)

class template:
    def __init__(self, prov: ModuleType) -> None:
        self.prov = prov
        ls = ["base_url", "rch_fn", "scanlator"]
        for i in ls:
            setattr(self, i, getattr(prov, i))
        if type(self.rch_fn) == str:
            self.rch_fn = partial(
                getattr(rch_fn, self.rch_fn),
                base_url=self.base_url,
                manga_id_fn=self.prov.manga_id_fn
            )
        self.template = import_module(f".{self.prov.template}", "mangdl.api.providers.templates").template

    def ch_fn(self, url: str) -> List[str]:
        src = getattr(self.prov, "src", "src")
        op = []
        for i in soup(url).select(".page-break img"):
            op.append(sanitize_text(i[src]))
        return op

    def chapter(self, url: str) -> Ch:
        ms = soup(url)
        return Ch(
            url              = url,
            ch               = self.prov.ch_num_fn(ms),
            vol              = None,
            title            = sanitize_text(ms.select_one("ol.breadcrumb .active").text.split("-")[-1]),
            scanlator_groups = [self.scanlator],
            imgs             = self.ch_fn(url),
        )

    def chdls(self, url: str, chs: int=0) -> List[Dict[Union[float, int, None], str]]:
        op = []
        rch = self.rch_fn(url)
        for c in rch.select("li.wp-manga-chapter    a"):
            cch = c["href"]
            if chs == 2:
                cch = self.chapter(cch)
            op.append({self.prov.rch_num_fun(c["href"]): cch})
        return op

    def manga(self, url: str, chs: int=0) -> Manga:
        ms = soup(url)
        meta = {}
        for c in ["post-content_item", "post-status"]:
            for m in ms.select(f"div.{c}"):
                v = m.select_one(".summary-content")
                idx = sanitize_text(m.select_one("h5").text)
                if v:
                    meta[idx] = sanitize_text(v.text)
                else:
                    meta[idx] = [sanitize_text(i.text) for i in m.select("a")]

        def mst(s: str, pp: Callable[[str], List[Any]]=lambda x: x):
            return pp(sanitize_text(meta.get(s))) if meta.get(s) else []

        rch = self.rch_fn(url)
        dates = [sanitize_text(i.text) for i in rch.select(".chapter-release-date")]
        return Manga(
            url             = url,
            covers          = [ms.select_one(".summary_image img")["src"]],
            title           = sanitize_text(ms.select_one("div.post-title h1").text),
            alt_titles      = mst("Alternative"),
            author          = mst("Author(s)", lambda x: x.split(",")),
            artist          = mst("Artist(s)", lambda x: x.split(",")),
            status          = {k: v for v, k in enumerate(["ongoing", "completed", "on hold"])}.get(meta["Status"].lower(), -1),
            genres          = mst("Genre(s)", lambda x: x.split(",")),
            updated_at      = dt(dates[0], r"%B %d, %Y") if dates else "1970-01-01T00:00:00",
            created_at      = dt(dates[-1], r"%B %d, %Y") if dates else "1970-01-01T00:00:00",
            chapters        = self.chdls(url, chs),
        )

    def dl_search(self, title: str, **kwargs: Dict[str, Any]):
        global total
        total = 0
        rpa = getattr(self.prov, "rpa", 12)
        cs_manga = getattr(self.prov, "cs_manga", "div.post-title")
        manga_check = getattr(self.prov, "manga_check", lambda x: True)
        title_fn = getattr(self.prov, "title_fn", lambda x: x.select_one("a").title)
        link_fn = getattr(self.prov, "link_fn", lambda x: x.select_one("a")["href"])
        sr = {}
        def pr(num: int):
            global total
            url = f"{self.base_url}/page/{num}?s={quote_plus(title)}&post_type=wp-manga"
            ms = soup(url)
            ts = ms.select(cs_manga)
            if not total:
                total = int(sanitize_text(ms.select_one(".c-blog__heading h1").text).split()[0])
            for r in ts:
                if manga_check(r):
                    sr[title_fn(r)] = link_fn(r)
            return len(ts)
        i = 1
        cp = pr(i)
        while cp // rpa and (rpa * i) < total:
            i += 1
            cp = pr(i)
        return sr

    def cli_search(self, title: str, **kwargs: Dict[str, Any]) -> Dict[str, str]:
        return self.dl_search(title, **kwargs)

    def dl(self, url: str, **kwargs: Dict[str, Any]):
        ms = soup(url)
        Downloader(self.ch_fn, **kwargs).dl_chdls(self.chdls(url), sanitize_text(ms.select_one("div.post-title h1").text))

    def cli_dl(self, title: str, **kwargs: Dict[str, Any]):
        Downloader(self.ch_fn, **kwargs).cli(self.template(self.prov).cli_search, partial(self.chdls), title)


from ast import literal_eval

from ..base import soup, urel

scanlator = "Mangasushi"
base_url = "https://mangasushi.net"
template = "wordpress"

src = "data-src"

def ch_num_fn(soup):
    return literal_eval(".".join(soup.select_one("#wp-manga-current-chap")["value"].split("-")[1:]))

def rch_num_fun(url):
    return literal_eval(".".join(urel(url).parts[3].split("-")[1:]))

def rch_fn(url):
    url = f"{base_url}/manga/{urel(url).parts[2]}/ajax/chapters/"
    return soup(url, method="post")


import json
from os import path
from os.path import abspath as ap
from os.path import dirname as dn
from typing import Any

import toml
import yaml


def readcfg(file: str) -> Any:
    """Read the contents of a file with the given file name.

    Args:
        file (str): File name of the file to read the contents of.

    Returns:
        Any: The contents of the file.
    """
    with open(file, "r") as f:
        d = {
            "json": lambda f: json.load(f),
            "yaml": lambda f: yaml.load(f.read(), yaml.SafeLoader),
            "toml": lambda f: toml.load(f)
        }
        return d[file.split(".")[-1]](f)


def stg(stg: str, file: str = path.join(dn(ap(__file__)), "stg.json")) -> Any:
    """Retrieve dictionary value of the config file with the given file name
    using recursive indexing with a string.
    ex.:
        Given that settings.json contains: `{"data": {"attr": {"ch": 1}}}`
        `stg("data/attr/ch", "settings.json")` will return `1`

    Args:
        stg (str): Directory of the value to be retrieved.
        file (str, optional): File name of the file to get the value from. Defaults to `path.join(dn(ap(__file__)), "settings.json")`.

    Returns:
        Any: The retrieved value.
    """
    op = readcfg(file)
    if stg is not None:
        for a in stg.split("/"):
            op = op[a]
    return op


def wr_stg(stg: str, value: Any, file: str = path.join(dn(ap(__file__)), "stg.json")) -> None:
    """Rewrite dictionary value of the config file with the given file name
    using recursive indexing with a string.
    ex.:
        Given that settings.json contains: `{"data": {"attr": {"ch": 1}}}`
        `wr_stg("data/attr/ch", 2)`
        will rewrite settings.json to be: `{"data": {"attr": {"ch": 2}}}`

    Args:
        stg (str): Directory of the value to be rewrited.
        value (Any): Value to rewrite to.
        file (str, optional): File name of the file to rewrite the value from. Defaults to path.join(dn(ap(__file__)), "settings.json").

    Raises:
        FileNotFoundError: Raised if the file is not found.
    """
    stg_dict = readcfg(file)

    def modify(stg: str, value: Any, stg_dict: dict[Any:Any]):
        path_ls = stg.split("/")
        key = path_ls[0]
        if len(path_ls) > 1:
            try:
                stg_dict[key]
            except KeyError:
                stg_dict[key] = {}
            if isinstance(stg_dict[key], dict):
                modify(stg.replace(f"{key}/", ""), value, stg_dict[key])
            else:
                f_stg = '"]["'.join(stg.split("/"))
                raise FileNotFoundError(f'["{f_stg}"] at {file} not found.')
        else:
            stg_dict[key] = value
            return stg_dict

    modify(stg, value, stg_dict)
    with open(path.join(dn(ap(__file__)), file), "w") as f:
        d = {
            "json": lambda f: json.dump(stg_dict, f, indent=4),
            "yaml": lambda f: yaml.dump(stg_dict, f, indent=4),
            "toml": lambda f: toml.dump(stg_dict, f)
        }
        d[file.split(".")[-1]](f)



from importlib import import_module
from typing import Any, Dict, List

from .api.base import Ch, Manga, Search


class Provider:
    def __init__(self, prov: str):
        self.prov = import_module(f".{prov}", "mangdl.api.providers")
        self.template = import_module(f".{self.prov.template}", "mangdl.api.providers.templates").template

    def chapter(self, url: str) -> Ch:
        local = locals()
        [local.pop(i) for i in ["self",]]
        return self.template(self.prov).chapter(
            **local,
        )

    def manga(self, url: str, chs: int=0) -> Manga:
        local = locals()
        [local.pop(i) for i in ["self",]]
        return self.template(self.prov).manga(
            **local,
        )

    def dl_search(self, title: str, **kwargs: Dict[str, Any]) -> Dict[str, str]:
        """Used for downloading when imported.
        Args:
            s (Search): Search dataclass, search parameters for searching.
        Returns:
            Dict[str, str]: Search results.
        """
        local = locals()
        [local.pop(i) for i in ["self",]]
        return self.template(self.prov).dl_search(
            **local,
        )

    def search(self, s: Search) -> List[Manga]:
        op = []
        for i in self.template(self.prov).dl_search(**s.__dict__).values():
            op.append(self.template.manga(i))
        return op

    def dl(self, url: str, **kwargs: Dict[str, Any]):
        local = locals()
        [local.pop(i) for i in ["self",]]
        return self.template(self.prov).dl(
            **local,
        )

    def cli_dl(self, title: str, **kwargs: Dict[str, Any]):
        local = locals()
        [local.pop(i) for i in ["self",]]
        return self.template(self.prov).cli_dl(
            **local,
        )


import builtins
import inspect
from typing import Any, Callable, List

import click
from tabulate import tabulate
from yachalk import chalk

from .providers import Provider
from .utils.settings import stg
from .utils.utils import dd, de, dnrp

print(chalk.hex("D2748D").bold(r"""
 _____   ___     _____    _____     __   _________   ______    __
/\    ＼/   \   / ____＼ /\    ＼  /\ \ /\  ______\ /\  ___＼ /\ \
\ \ \＼   /\ \ /\ \__/\ \\ \ \＼ ＼\ \ \\ \ \_____/_\ \ \_/\ \\ \ \
 \ \ \ ＼/\ \ \\ \  ____ \\ \ \ ＼ ＼ \ \\ \ \ /\__ \\ \ \\ \ \\ \ \
  \ \ \‾‾  \ \ \\ \ \__/\ \\ \ \  ＼ ＼\ \\ \ \\/__\ \\ \ \\_\ \\ \ \____
   \ \_\    \ \_\\ \_\ \ \_\\ \_\＼ ＼____\\ \________\\ \_____/ \ \_____\
    \/_/     \/_/ \/_/  \/_/ \/_/  ＼/____/ \/________/ \/____/   \/_____/

The most inefficient, non user-friendly and colorful manga downloader (and soon, also a reader)""") + chalk.hex("3279a1")('\nChat with whi~nyaan at https://discord.com/invite/JbAtUxGcJZ\n'))

def cao(group: click.group, cmd: str) -> List[Callable[[Callable[[Any], Any]], Callable[[Any], Any]]]:
    """Retruns wrappers for a click command evaluated from the given arguments.

    Args:
        group (click.group): Command group of the command to be under.
        cmd (str): Name of the command.

    Returns:
        List[Callable[[Callable[[Any], Any]], Callable[[Any], Any]]]: The wrappers.
    """
    cmd = stg(f"cmd/{cmd}", f"{dnrp(__file__)}/utils/config.yaml")
    arguments = cmd["arguments"]

    def c(f: Callable[[Any], Any]) -> Callable[[Callable[[Any], Any]], Callable[[Any], Any]]:
        """The command wrapper.
        Args:
            f (Callable[[Any], Any]): The command function to be decorated.
        Returns:
            Callable[[Callable[[Any], Any]], Callable[[Any], Any]]
        """
        help = []
        if arguments:
            for k, v in arguments.items():
                arguments[k]["help"] = [*v["help"], *[None for _ in range(3 - len(v["help"]))]]
            for k, v in arguments.items():
                t, h, e = v["help"]
                e = '\nEx.: {e}' if e else ""
                help.append([f"<{k}>", t, f'{h}{e}'])
        s, h = cmd["help"]
        return group.command(*de(cmd["args"], []), **dd({"context_settings": {'help_option_names': ['-h', '--help']}, "short_help": s, "help": f"\b\n{h}\n{tabulate(help, tablefmt='plain')}"}, cmd["kwargs"]))(f)

    def a(f: Callable[[Any], Any]) -> Callable[[Callable[[Any], Any]], Callable[[Any], Any]]:
        """The arguments wrapper.
        Args:
            f (Callable[[Any], Any]): The command function to be decorated.
        Returns:
            Callable[[Callable[[Any], Any]], Callable[[Any], Any]]
        """
        args = {}
        kwargs = {}
        if arguments:
            for k, v in arguments.items():
                kw = {"metavar": f"<{k}>"}
                args[k] = [k, *de(v["args"], [])]
                kwargs[k] = dd(kw, v["kwargs"])
            for i in list(args.keys()):
                f = click.argument(*args[i], **kwargs[i])(f)
        return f

    def o(f: Callable[[Any], Any]) -> Callable[[Callable[[Any], Any]], Callable[[Any], Any]]:
        """The options wrapper.
        Args:
            f (Callable[[Any], Any]): The command function to be decorated.
        Returns:
            Callable[[Callable[[Any], Any]], Callable[[Any], Any]]
        """
        if opts:= cmd["options"]:
            n = 0
            args = {}
            kwargs = {}
            for k, v in opts.items():
                l = len(v["help"][0] or "")
                n = l if l > n else n
                opts[k]["help"] = [*v["help"], *[None for _ in range(3 - len(v["help"]))]]
            for k, v in opts.items():
                a = de(v["args"], [])
                kw = de(v["kwargs"], {})
                a[0] = f"--{a[0]}"
                a.insert(0, f"-{k}")
                kt = kw.get("type", None)
                t, h, e = v["help"]
                t = t or ""
                h = "\n".join(f'{" " * ((n + 3)-(0 if i else len(j)))}{j}' for i, j in enumerate(h.split("\n"))) if h else ""
                e = "\n" + "\n".join(f'{" " * (n + (11 if i else 5))}{"Ex.: " if not i else ""}{j}' for i, j in enumerate(e.split("\n"))) if e else ""
                kw["help"] = f'\b\n{t}{" "*((n + 3) - len(t))}{h}{e}'
                if type(kt) is dict:
                    ktk, ktv = list(kt.items())[0]
                    kta, ktkw = [i[1] for i in ktv.items()]
                    kw["type"] = getattr(click, ktk)(*kta, **ktkw if ktkw else {})
                elif kt:
                    kw["type"] = getattr(builtins, kt)
                args[k] = a
                kwargs[k] = kw
            for i in list(args.keys()):
                f = click.option(*args[i], **kwargs[i])(f)
        return f

    return c, a, o

def command(group: click.group) -> Callable[[Callable[[Any], Any]], Callable[[Any], Any]]:
    """Wrapper for click commands.

    Args:
        group (click.group): Command group of the command to be under.

    Returns:
        Callable[[Callable[[Any], Any]], Callable[[Any], Any]]
    """
    def inner(f: Callable[[Any], Any]):
        m = inspect.getouterframes(inspect.currentframe())[1][4][0]
        for m in cao(group, m[4:m.index("(")]):
            f = m(f)
        return f
    return inner

@click.group(context_settings={'help_option_names': ['-h', '--help']})
def cli():
    """
    Main command group.
    """
    pass

@command(cli)
def dl(title: str, **kwargs: dict[str, Any]):
    """Download command.

    Args:
        title (str): The title of the manga to be search for and download.
    """
    Provider(kwargs.pop("provider", "mangadex")).cli_dl(title, **kwargs)

@command(cli)
def credits():
    """Credits command. Display credits.
    """
    print(chalk.hex("D2748D").bold(r"""
 ______  ______  ______  _____   __  ______  ______
/\  ___\/\  == \/\  ___\/\  __-./\ \/\__  _\/\  ___\
\ \ \___\ \  __<\ \  __\\ \ \/\ \ \ \/_/\ \/\ \___  \
 \ \_____\ \_\ \_\ \_____\ \____-\ \_\ \ \_\ \/\_____\
  \/_____/\/_/ /_/\/_____/\/____/ \/_/  \/_/  \/_____/""") + chalk.hex("11b180").bold("\n\nThank you:") + chalk.hex("3279a1")("""\n
- To Arjix, who helped me in implementing majority of the features and de-minifying
  my code, making it more readable and more efficient at the same time
- To KR, who let me use the KR-naming scheme like "AnimDL" do
- To whi~nyaan, my alter ego, for just existing (and purring, ofc)
- And to everyone who supported me from the very beginning of this humble
  project to its release!\n"""))



from functools import partial
from importlib import import_module
from types import ModuleType
from typing import Any, Callable, Dict, List, Union
from urllib.parse import quote_plus

from bs4 import BeautifulSoup

from ....utils.utils import dt, sanitize_text
from ...base import Ch, Downloader, Manga, soup


class template:
    def __init__(self, prov: ModuleType) -> None:
        self.prov = prov
        ls = ["base_url", "scanlator"]
        for i in ls:
            setattr(self, i, getattr(prov, i))
        self.check_manga = getattr(self.prov, "check_manga", lambda x: True)
        self.template = import_module(f".{self.prov.template}", "mangdl.api.providers.templates").template

    def ch_fn(self, url: str) -> List[str]:
        return [i["src"] for i in soup(url).select("#readerarea p img")]

    def chapter(self, url: str) -> Ch:
        ch = soup(url).select_one("h1.entry-title").text.split()[-1]
        return Ch(
            url              = url,
            ch               = ch,
            vol              = None,
            title            = f"Chapter {ch}",
            scanlator_groups = [self.scanlator],
            imgs             = self.ch_fn(url),
        )

    def chdls(self, url: str, chs: int=0) -> List[Dict[Union[float, int, None], str]]:
        op = []
        for c in soup(url).select("div.eplister ul li"):
            cch = c.select_one("a")["href"]
            if chs == 2:
                cch = self.chapter(cch)
            op.append({c["data-num"]: cch})
        return op

    def manga(self, url: str, chs: int=0) -> Manga:
        ms = soup(url)
        meta = {}
        for c in ["post-content_item", "post-status"]:
            for m in ms.select(f"div.{c}"):
                v = m.select_one(".summary-content")
                idx = sanitize_text(m.select_one("h5").text)
                if v:
                    meta[idx] = sanitize_text(v.text)
                else:
                    meta[idx] = [sanitize_text(i.text) for i in m.select("a")]

        def mst(s: str, pp: Callable[[str], List[Any]]=lambda x: x):
            return pp(sanitize_text(meta.get(s))) if meta.get(s) else []

        rch = self.prov.rch_fn(url)
        dates = [sanitize_text(i.text) for i in rch.select(".chapter-release-date")]
        return Manga(
            url             = url,
            covers          = [ms.select_one(".summary_image img")["src"]],
            title           = sanitize_text(ms.select_one("div.post-title h1").text),
            alt_titles      = mst("Alternative"),
            author          = mst("Author(s)", lambda x: x.split(",")),
            artist          = mst("Artist(s)", lambda x: x.split(",")),
            status          = {k: v for v, k in enumerate(["ongoing", "completed", "on hold"])}.get(meta["Status"].lower(), -1),
            genres          = mst("Genre(s)", lambda x: x.split(",")),
            updated_at      = dt(dates[0], r"%B %d, %Y") if dates else "1970-01-01T00:00:00",
            created_at      = dt(dates[-1], r"%B %d, %Y") if dates else "1970-01-01T00:00:00",
            chapters        = self.chdls(url, chs),
        )

    def manga(self, url: str, chs: int=0) -> Manga:
        ms = soup(url)
        meta = {}
        for m in ms.select(".tsinfo.bixbox .imptdt"):
            i = m.select_one("i")
            meta[m.next_element.strip()] = i.text if i else m.select_one("a").text

        return Manga(
            url             = url,
            covers          = [ms.select_one(".thumb .attachment-.size-.wp-post-image")["src"]],
            title           = ms.select_one("h1.entry-title").text,
            alt_titles      = ms.select_one("span.alternative").text.split(" | "),
            author          = meta["Author"].split(", "),
            status          = {k: v for v, k in enumerate(["Ongoing", "Completed", "Hiatus", "Cancelled"])}.get(meta["Status"], -1),
            updated_at      = dt(meta["Updated On"], r"%B %d, %Y"),
            created_at      = dt(meta["Posted On"], r"%B %d, %Y"),
            description     = "\n".join(i.text for i in ms.select(".entry-content.entry-content-single p")),
            chapters        = self.chdls(url, chs),
        )

    def dl_search(self, title: str, **kwargs: Dict[str, Any]) -> Dict[str, str]:
        sr = {}
        def pr(soup: BeautifulSoup):
            for r in soup.select(".listupd .bs .bsx a"):
                if self.check_manga(r):
                    sr[r["title"]] = r["href"]
        ms = soup(f"{self.base_url}/?s={quote_plus(title)}")
        pr(ms)
        pages = ms.select("a.page-numbers")
        if pages:
            for p in range(int(pages[-2].text)-1):
                pr(soup(f"{self.base_url}/page/{p+2}/?s={quote_plus(title)}"))
        return sr

    def cli_search(self, title: str, **kwargs: Dict[str, Any]) -> Dict[str, str]:
        return self.dl_search(title, **kwargs)

    def dl(self, url: str, **kwargs: Dict[str, Any]):
        ms = soup(url)
        Downloader(self.ch_fn, **kwargs).dl_chdls(self.chdls(url), sanitize_text(ms.select_one("div.post-title h1").text))

    def cli_dl(self, title: str, **kwargs: Dict[str, Any]):
        Downloader(self.ch_fn, **kwargs).cli(self.template(self.prov).cli_search, partial(self.chdls), title)


import ast
import difflib
import itertools
import re
import sys
import unicodedata
from datetime import datetime, timedelta
from functools import lru_cache
from os.path import dirname as dn
from os.path import realpath as rp
from time import strftime, strptime
from typing import Any, Callable, Dict, List, Union

# https://stackoverflow.com/a/93029
ALL_CHARS = (chr(i) for i in range(sys.maxunicode))
CATEGORIES = {'Cn'}
CCHARS = ''.join(map(chr, itertools.chain(range(0x00,0x20), range(0x7f,0xa0))))
CCHARS_RE = re.compile('[%s]' % re.escape(CCHARS))

def dnrp(file: str, n: int=1) -> str:
    """
    Get the directory component of a pathname by n times recursively then return it.

    Args:
        file (str): File to get the directory of.
        n (int, optional): Number of times to get up the directory???? Defaults to 1.

    Returns:
        op (str): The directory component got recursively by n times from the given pathname
    """
    op = rp(file)
    for _ in range(n):
        op = dn(op)
    return op

def de(a: Any, d: Any) -> Any:
    """
    Defaults. Return a if a is True, else returns d.

    Args:
        a (Any): Object to be tested, will be returned if evaluates to True.
        d (Any): Default object to be returned if `a` evaluates to False.

    Returns:
        Any
    """
    if a:
        return a
    else:
        return d

def dd(default: Dict[Any, Any], d: Union[Dict[Any, Any], None]) -> Dict[Any, Any]:
    """Defaults dictionary. Overwrite the items in the default dict with the
    items in the d dict.

    Args:
        default (Dict[Any, Any]): The dict to rewrite the items to.
        d (Union[Dict[Any, Any], None]): The dict to rewrite the items from.

    Returns:
        dict[Any, Any]
    """
    op = default
    if d:
        for a, v in d.items():
            op[a] = v
    return op

def ddir(d: Dict[Any, Any], dir: str) -> Any:
    """Retrieve dictionary value using recursive indexing with a string.
    ex.:
        `ddir({"data": {"attr": {"ch": 1}}}, "data/attr/ch")`
        will return `1`


    Args:
        dict (dict): Dictionary to retrieve the value from.
        dir (str): Directory of the value to be retrieved.

    Returns:
        op (Any): Retrieved value.
    """
    op = d
    for a in dir.split("/"):
        op = op[a]
    return op

@lru_cache
def dt(dt: str, format: str) -> str:
    """
    Remove timezone from datetime and format it to ISO 8601 format.

    Args:
        dt (str): Unformatted datetime string to be formatted to ISO 8601 format
        format (str): The initial format of the datetime string

    Returns:
        str: Formatted datetime string
    """
    dt = strftime("%Y-%m-%dT%H:%M:%S", strptime(dt, format))
    tz = re.match(r"(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2})([-+])(\d{2}):(\d{2})", dt)
    if tz:
        iso, s, ho, mo = tz.groups()
        s = -1 if s == "-" else 1
        return (datetime.fromisoformat(iso) - (s * timedelta(hours=int(ho), minutes=int(mo)))).strftime("%Y-%m-%dT%H:%M:%S")
    else:
        return dt

@lru_cache
def dt_ts(ts: str) -> str:
    """
    Convert the given unix timestamp to ISO 8601 format.

    Args:
        ts (str): unix timestamp to be converted to ISO 8601 format

    Returns:
        str: Formatted datetime string
    """
    return (datetime.utcfromtimestamp(int(ts))).strftime("%Y-%m-%dT%H:%M:%S")

def parse_list(opt: Union[str, None]):
    """
    Takes a string and evaluates it to a list if it is a list,
    else it splits it into a list.

    It then evaluates the values of that list.

    Args:
        opt (Union[str, None]): The object to be evaluated.
    """

    if type(opt) is str:
        opt = opt[1:-1]
        try:
            opts = list(ast.literal_eval(opt))
        except (SyntaxError, ValueError):
            opts = list(opt.split(','))
        for i, x in enumerate(opts):
            try:
                opts[i] = ast.literal_eval(x)
            except (SyntaxError, ValueError):
                opts[i] = x.strip()
        return opts
    else:
        return None

def squery(query: str, possibilities: List[str], cutoff: int=0.6, *, processor: Callable[[Any], Any]=lambda r: r):
    """Custom search query.

    Args:
        query (str): String to search for in the possibilities.
        possibilities (List[str]): The possibilities to search from.
        cutoff (int, optional): The minimum percentage of similarity from the
            given possibilities. Defaults to 0.6.
        processor (Callable[[Any], Any], optional): Processes the possibilities
            before comparing it with the query. Defaults to lambda r:r.
    """

    sequence_matcher = difflib.SequenceMatcher()
    sequence_matcher.set_seq2(query)

    for search_value in possibilities:
        sequence_matcher.set_seq1(processor(search_value))
        if (query.lower() in processor(search_value).lower()):
            yield (None, search_value)
            continue
        if (sequence_matcher.real_quick_ratio() >= cutoff and sequence_matcher.quick_ratio(
        ) >= cutoff and sequence_matcher.ratio() >= cutoff):
            yield (sequence_matcher.ratio(), search_value)

def sanitize_text(s: str):
    return unicodedata.normalize("NFKD", CCHARS_RE.sub('', s)).strip()


from types import ModuleType
from typing import Any, Dict, List, Union

from yarl import URL

from ....utils.utils import ddir, dt, dt_ts, squery
from ...base import Ch, Manga, Search, ddos_guard, req, soup, urel


class template:
    def __init__(self, prov: ModuleType) -> None:
        self.prov = prov
        ls = ["base_url", "scanlator"]
        for i in ls:
            setattr(self, i, getattr(prov, i))
        self.links_fn = getattr(self.prov, "links_fn", lambda x: {})

    def api_series(self, url: str):
        url = f"{self.base_url}/api/series/{urel(url).parts[3]}/"
        return req.get(url).json()

    def ch_fn(self, url: str) -> List[str]:
        url = urel(url)
        resp = self.api_series(url)
        meta = ddir(resp, f"chapters/{url.parts[4]}")
        bui = f'{self.base_url}/media/manga/{url.parts[3]}/chapters/{meta["folder"]}/1/'
        return [bui + i for i in ddir(meta, "groups/1")]

    def chapter(self, url: str) -> Ch:
        resp = self.api_series(url)
        url = urel(url)
        slug = url.parts[3]
        ch = int(url.parts[4])
        meta = ddir(resp, f"chapters/{ch}")
        ms = soup(f'{self.base_url}/read/manga/{slug}', ddos_guard)
        bui = f'{self.base_url}/media/manga/{slug}/chapters/{meta["folder"]}/1/'
        return Ch(
            url              = url,
            ch               = ch,
            vol              = meta["volume"],
            title            = meta["title"],
            views            = int(ms.select_one(f'tr[data-chapter="{ch}"] .text-right').text),
            uploaded_at      = dt_ts(list(ddir(meta, "release_date").values())[0]),
            scanlator_groups = list(resp['groups'].values()),
            user             = [self.scanlator],
            imgs             = [bui + i for i in list(ddir(meta, "groups").values())[0]],
        )

    def chdls(self, url: str, chs: int=0) -> List[Dict[Union[float, int, None], str]]:
        meta = self.api_series(url)
        op = []
        for c in meta["chapters"]:
            cch = f"{self.base_url}/read/manga/{urel(url).parts[3]}/{c}/"
            if chs == 2:
                cch = self.chapter(cch)
            op.append({c: cch})
        return op

    def manga(self, url: str, chs: int=0) -> Manga:
        meta = self.api_series(url)
        slug = urel(url).parts[3]
        ms = soup(f'{self.base_url}/read/manga/{slug}', ddos_guard)

        mls = [i.text for i in ms.select('.col-lg-8 table.table-borderless th')]
        def rmh(meta: str):
            idx = mls.index(meta) + 1
            cs = f'.col-lg-8 table.table-borderless tr:nth-child({idx}) td'
            return ms.select_one(cs)
        def mh(meta: str):
            return rmh(meta).text

        return Manga(
            url             = str(URL.build(scheme="https", host="danke.moe", path=urel(url).path)),
            covers          = [meta["cover"]],
            title           = meta["title"],
            author          = [meta["author"]],
            created_at      = dt(mh("Last Updated").split(" - ")[1], "%Y-%m-%d"),
            views           = int(mh("Views")),
            description     = ms.select_one("div.col-md-7 > article > p").text,
            links           = self.links_fn(rmh),
            chapters        = self.chdls(url, chs),
        )

    def dl_search(self, title: str, **kwargs: Dict[str, Any]) -> Dict[str, str]:
        ms = soup(f"{self.base_url}/series", ddos_guard)
        all_series = {i.text: i["href"] for i in ms.select("h7.card-title a")}
        sr = {}
        for series, series_url in all_series.items():
            if list(squery(title, [series])):
                sr[series] = self.base_url + series_url
        return sr

    def search(self, s: Search) -> List[Manga]:
        return [self.manga(i) for i in self.dl_search(s).values()]

    def cli_search(self, title: str, **kwargs: Dict[str, Any]):
        return self.dl_search(title, **kwargs)


import ast
import os
import re
import shutil
import sys
import time
from dataclasses import dataclass, field
from functools import partial
from multiprocessing.pool import ThreadPool
from typing import Any, Callable, Dict, List, Type, Union

import click
import httpx
import patoolib
from bs4 import BeautifulSoup
from tabulate import tabulate
from tqdm import tqdm
from yachalk import chalk
from yarl import URL

from ..utils import style
from ..utils.exceptions import DownloadFailed

METHODS = ["get", "options", "head", "post", "put", "patch", "delete"]
SESSION = httpx.Client()

def _req(
        url: str,
        ra: Callable[[httpx.Response], int]=None,
        method: str = "get",
        session: httpx.Client = SESSION,
        *args: List[Any],
        **kwargs: Dict[str, Any]
    ) -> httpx.Response:
    """Custom request function with retry after capabilities for 429s.

    Args:
        url (str): URL to send the request to.
        ra (Callable[[httpx.Response], int], optional): Retry after function,
            receives the Response object and returns the seconds before
            retrying. Defaults to None.
        method (str, optional): The request method. Defaults to "get".
        session (httpx.Client, optional): Session client. Defaults to SESSION.

    Returns:
        httpx.Response: Response object.
    """
    resp = getattr(session, method)(url, follow_redirects=True, *args, **kwargs)
    if resp.status_code == 429:
        if ra:
            time.sleep(ra(resp))
            return _req(url, ra, method, session, *args, **kwargs)
    return resp

class req:
    pass

class ddos_guard:
    pass

ddos_resp = httpx.post("https://check.ddos-guard.net/check.js")
ddos_cookies = {k: v for k, v in ddos_resp.cookies.items()}
ddos_session = httpx.Client(cookies=ddos_cookies)

req_dict = {
    req: {},
    ddos_guard: {"session": ddos_session},
}

for r, kw in req_dict.items():
    for i in METHODS:
        setattr(r, i, partial(_req, method=i, **kw))

def soup(url: str, req: Type[req]=req, method: str="get", **kwargs: Dict[str, Any]) -> BeautifulSoup:
    """Returns a soup from the given url.

    Args:
        url (str): URL to get the soup from.
        req (Type[req], optional): Object to call the methods from. Defaults to req.

    Returns:
        BeautifulSoup: the soup
    """
    return BeautifulSoup(getattr(req, method)(url, **kwargs).text, "lxml")

@dataclass
class Vls:
    vdict   : Dict[str, Any]
    ls      : List[str]

@dataclass
class Search:
    title           : str
    lang            : List[str]                             = field(default_factory=list)
    excludelang     : List[str]                             = field(default_factory=list)
    demo            : List[str]                             = field(default_factory=list)
    contentrating   : List[str]                             = field(default_factory=list)
    status          : List[str]                             = field(default_factory=list)
    order           : str                                   = field(default_factory=dict)
    authors         : List[str]                             = field(default_factory=list)
    artists         : List[str]                             = field(default_factory=list)
    year            : int                                   = None
    includetags     : List[str]                             = field(default_factory=list)
    includemode     : str                                   = None
    excludetags     : List[str]                             = field(default_factory=list)
    excludemode     : str                                   = None
    range           : Callable[[Union[int, float]], bool]   = lambda *args, **kwargs: True
    cover           : bool                                  = False
    directory       : str                                   = None
    overwrite       : bool                                  = False
    retry           : int                                   = 3
    retryprompt     : bool                                  = False
    threads         : int                                   = 5
    verbosity       : int                                   = 4
    saveconfig      : str                                   = None
    loadconfig      : str                                   = None
    overridelc      : bool                                  = False


@dataclass
class Ch:
    url              : str
    ch               : Union[int, float, None]
    vol              : Union[int, float, None]
    title            : str
    views            : int                      = 0
    uploaded_at      : str                      = None
    scanlator_groups : List[str]                = field(default_factory=list)
    user             : str                      = None
    imgs             : List[str]                = field(default_factory=list)


@dataclass
class Manga:
    url             : str
    title           : str
    author          : List[str]
    artist          : List[str]                     = field(default_factory=list)
    covers          : List[str]                     = field(default_factory=list)
    alt_titles      : List[str]                     = field(default_factory=list)
    status          : int                           = -1
    demographics    : str                           = None
    content_rating  : str                           = None
    genres          : Dict[str, str]                = field(default_factory=dict)
    updated_at      : str                           = None
    created_at      : str                           = None
    views           : int                           = None
    description     : str                           = "No description available."
    links           : Dict[str, str]                = field(default_factory=dict)
    chapters        : Dict[Union[int, float], Ch]   = field(default_factory=list)


def tblp(ls: List[str], ct: str="title", prompt: str='Enter the index of the manga to be downloaded'):
    """Table prompt.
    Receive a list of items, format it in a table form and print,
    then prompt the user to choose from the list using an index.

    Args:
        ls (List[str]): List of items.
        ct (str, optional): [description]. Defaults to "title".
        prompt (str, optional): [description]. Defaults to 'Enter the index of the manga to be downloaded'.

    Returns:
        click.prompt: The input of the user.
    """
    print(
        tabulate(
            [
                [chalk.hex("FFD166").bold(i), chalk.hex("4E8098").bold(v)]
                for i, v in enumerate(ls)
            ],
            [chalk.hex("e63946").bold("index"), chalk.hex("e76f51").bold(ct)],
            tablefmt="pretty", colalign=("right", "left")
        )
    )

    return click.prompt(
        chalk.hex("3279a1").bold(
            f'{prompt}, defaults to 0'
        ), '0',
        type=click.Choice(
            [str(i) for i in range(len(ls))]
        ),
        show_choices=False,
        show_default=False
    )


def sanitize_filename(filename: str) -> str:
    """Sanitize the given filename.

    Args:
        filename (str): The filename to be sanitized.

    Returns:
        str: Sanitized filename.
    """
    return re.sub(re.compile(r"[<>:\"/\\|?*]", re.DOTALL), "", str(filename))


def get_extension(filename: str) -> str:
    """Get the file extension of a file from the given filename.

    Args:
        filename (str): The filename to get the file extension from.

    Returns:
        str: The file extension from the given filename.
    """
    return filename.strip("/").split("/")[-1].split("?")[0].split(".")[-1]

def ordinal(n: int) -> str:
    """Convert the given number to ordinal number.

    Args:
        n (int): The number to convert into ordinal number.

    Returns:
        str: The said ordinal number.
    """
    return "%d%s" % (n,"tsnrhtdd"[(n//10%10!=1)*(n%10<4)*n%10::4])

def _cr(rs: str):
    """The function that calculates the range.

    Args:
        rs (str): The range string where the range is calculated from.

    Yields:
        Callable[[Any], bool]: The function that checks if the given int is within the range or not.
    """

    for matches in re.finditer(r"(?:(-?[0-9.]*)[:](-?[0-9.]*)|(-?[0-9.]+))", rs):
        start, end, singular = matches.groups()
        if (start and end) and float(start) > float(end):
            start, end = end, start
        yield (lambda x, s=singular: float(s) == x) if singular else (lambda x: True) if not (start or end) else (lambda x, s=start: x >= float(s)) if start and not end else (lambda x, e=end: x <= float(e)) if not start and end else (lambda x, s=start, e=end: float(s) <= x <= float(e))
    if not rs:
        return lambda *args, **kwargs: True

def cr(rs: str) -> Callable[[int], bool]:
    """Returns a function that checks if the given int is within the range or not.
    The range is calculated from the given string.

    Args:
        rs (str): The range string where the range is calculated from.

    Returns:
        Callable[[int], bool]: The function that checks if the given int is within the range or not.
    """


    if rs:
        return lambda x: any(condition(x) for condition in _cr(rs))
    else:
        return lambda x: True


class Downloader:
    def __init__(
        self,
        ch_fn: Callable[[Any], List[str]],
        ra: Callable[[httpx.Response], int]=None,
        headers: Dict[str, Any] = {},
        range: str='',
        directory: str=None,
        overwrite: bool=True,
        format: str='cbz',
        delfolder: bool=True,
        retry: int=3,
        retryprompt: bool=False,
        threads: int=30,
        **kwargs: Dict[str, Any]
    ):
        local = locals()
        for i in ["ch_fn", "ra", "headers", "overwrite", "delfolder", "retry", "retryprompt", "threads"]:
            setattr(self, i, local[i])
        if ddir:= directory:
            self.ddir = ddir
        elif sys.platform == "win32":
            self.ddir = os.path.join(os.path.join(os.environ['USERPROFILE']), 'Desktop', 'Manga')
        else:
            self.ddir = os.path.join(os.path.expanduser('~'), "Manga")
        self.format = format.lower()
        self.check = cr(range)

    def _dlf(self, file: list[str], n: int=0):
        """The core individual image downloader.

        Args:
            file (str): List containing the filename and the url of the file.
            n (int, optional): Times the download for this certain file is retried. Defaults to 0.
        """
        if (n+1) == self.retry:
            return False
        else:
            with open(file[0] + ".tmp", "wb") as f:
                try:
                    with httpx.stream("GET", file[1], headers=self.headers) as r:
                        if r.status_code == 200:
                            for chunk in r.iter_bytes(chunk_size=8192):
                                f.write(chunk)
                        elif r.status_code == 429:
                            retry = self.ra(r)
                            time.sleep(retry)
                            self._dlf(file, n)
                except (httpx.ReadTimeout, httpx.ConnectTimeout):
                    self._dlf(file, n+1)
            return True

    def dlf(self, file: List[str]):
        """Individual image downloader.

        Args:
            file (str): List containing the filename and the url of the file.
        """
        try:
            os.makedirs(os.path.split(file[0])[0])
        except FileExistsError:
            pass
        if os.path.isfile(f"{file[0]}.tmp"):
            os.remove(f"{file[0]}.tmp")
        f = self._dlf(file)
        if f:
            os.replace(f"{file[0]}.tmp", file[0])
        else:
            raise DownloadFailed(f"Download of {file[1]} failed.")

    def dlch(self, k: Union[int, float], v: List[str], n: int=0):
        """Individual chapter downloader.

        Args:
            k (Union[int, float]): Chapter number.
            v (List[str]): List of image urls.
        """
        dl = True
        jdir = os.path.join(self.ddir, self.title, sanitize_filename(k))
        if self.check(k):
            if os.path.isdir(jdir):
                if self.overwrite:
                    shutil.rmtree(jdir)
                else:
                    dl = False
            if os.path.isfile(f"{jdir}.{self.format}"):
                if self.overwrite:
                    os.remove(f"{jdir}.{self.format}")
                else:
                    dl = False
        if (n+1) == self.retry:
            dl = False
        if dl:
            chapter_name = str(k)
            jdir = os.path.join(self.ddir, self.title, sanitize_filename(chapter_name))
            dl = True
            if dl:
                files = []
                for index, page in enumerate(self.ch_fn(v)):
                    filename = os.path.join(
                        self.ddir,
                        self.title,
                        sanitize_filename(chapter_name),
                        f"{index}.{get_extension(str(page))}"
                    ).replace("\\", "/")
                    files.append((filename, page))
                fmt = style.t1(style.ac1(chapter_name) + " [{remaining_s:05.2f} secs, {rate_fmt:0>12}] " + style.ldb("{bar}") +" [{n:03d}/{total:03d}, {percentage:03.0f}%]")
                try:
                    with ThreadPool(self.threads) as pool:
                        list(tqdm(pool.imap(self.dlf, files), total=len(files), leave=True, unit=" img", disable=False, dynamic_ncols=True, smoothing=1, bar_format=fmt))
                except DownloadFailed as e:
                    if self.retryprompt:
                        self.dlch(k, v, n+1)
                    else:
                        raise e

                if self.format != "folder":
                    patoolib.create_archive(f"{jdir}.{self.format}", [i[0] for i in files], verbosity=-1)
                    if self.delfolder:
                        shutil.rmtree(jdir)

    def dl(self, title: str, chs: Dict[Union[int, float], list[str]]):
        self.title = title
        for k, v in chs.items():
            self.dlch(k, v)

    def dl_chdls(
        self,
        chdls: List[Dict[Union[float, int, None], str]],
        title: str,
    ):
        n = 1
        chs = {}
        for r in chdls:
            ch, arg = list(r.items())[0]
            if ch:
                if type(ch) is int or type(ch) is float:
                    k = ch
                else:
                    k = ast.literal_eval(ch)
            else:
                k = -n
                n += 1
            chs[k] = arg
        self.dl(title, chs)

    def cli(
        self,
        cli_search: Callable[[str], Dict[str, str]],
        chdls: Callable[[str], List[Dict[Union[float, int, None], str]]],
        title: str,
    ):
        sr = cli_search(title)
        if sr:
            ls = list(sr.keys())
            title = ls[int(tblp(ls))]
            self.dl_chdls(chdls(sr[title]), title)
        else:
            print(style.warning(f"No manga with similar title with the requested title or anything similar found. Use other search terms or remove some filters."))

def urel(url: str):
    url = URL(url)
    if url.is_absolute():
        url = url.relative()
    return url


from ast import literal_eval

from mangdl.utils.utils import sanitize_text

from ..base import urel

scanlator = "Paean Scans"
base_url = "https://setsuscans.com"
template = "wordpress"

cs_manga = "div.tab-summary"
rch_fn = "setsu"
src = "data-src"

def ch_num_fn(soup):
    return literal_eval(soup.select_one("#chapter-heading").text.split("-")[-1][9:])

def rch_num_fun(url: str):
    return literal_eval(".".join(urel(url).parts[3].split("-")[1:]))

def manga_id_fn(soup) -> str:
    return soup.select_one("#manga-chapters-holder")["data-id"]

def manga_check(soup):
    return not sanitize_text(soup.select_one(".mg_genres a").text) == "Bilibili"

def title_fn(soup):
    return soup.select_one(".post-title a").text

def link_fn(soup):
    return soup.select_one(".post-title a")["href"]


from types import ModuleType
from typing import Any, Dict, List, Union

from ...base import Ch, Manga


class template:
    def __init__(self, prov: ModuleType) -> None:
        self.prov = prov

    def ch_fn(self, url: str) -> List[str]:
        local = locals()
        [local.pop(i) for i in ["self",]]
        return self.prov.ch_fn(**local)

    def chapter(self, url: str) -> Ch:
        local = locals()
        [local.pop(i) for i in ["self",]]
        return self.prov.chapter(**local)

    def chdls(self, url: str, chs: int=0) -> List[Dict[Union[float, int, None], str]]:
        local = locals()
        [local.pop(i) for i in ["self",]]
        return self.prov.chdls(**local)

    def manga(self, url: str, chs: int=0) -> Manga:
        local = locals()
        [local.pop(i) for i in ["self",]]
        return self.prov.manga(**local)

    def dl_search(self, title: str, **kwargs: Dict[str, Any]):
        return self.prov.dl_search(title, **kwargs)

    def cli_search(self, title: str, **kwargs: Dict[str, Any]) -> Dict[str, str]:
        local = locals()
        [local.pop(i) for i in ["self",]]
        return self.prov.clu_search(**local)

    def dl(self, url: str, **kwargs: Dict[str, Any]):
        local = locals()
        [local.pop(i) for i in ["self",]]
        return self.prov.dl(**local)

    def cli_dl(self, title: str, **kwargs: Dict[str, Any]):
        local = locals()
        [local.pop(i) for i in ["self",]]
        return self.prov.cli_dl(**local)


from yachalk import chalk

from .settings import stg
from .utils import dnrp

cstyle = stg(stg("color_theme"), f"{dnrp(__file__)}/colours.yaml")
t1 = chalk.hex(cstyle["t1"])
ldb = chalk.hex(cstyle["fg1"]).bg_hex(cstyle["bg1"])
for i in [*[f"t{i}" for i in range(2, 8)], "ac1", 'critical', 'error', 'warning', 'info', 'debug']:
    globals()[i] = chalk.hex(cstyle[i]).bold



from setuptools import find_packages, setup

with open("version", "r") as f:
    version = f.read().strip()

setup(
    name="MangDL",
	author="whinee",
	author_email="whinyaan@gmail.com",
    version=version,
    description="The most inefficent Manga downloader for PC",
    long_description="file: docs/README.md",
    long_description_content_type="text/markdown",
    url="https://github.com/MangDL/MangDL",
	license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
	packages=find_packages(),
	include_package_data=True,
    python_requires=">=3.9",
	install_requires=[
        "BeautifulSoup4",
		"click",
        "httpx",
		"lxml",
        "patool",
        "pyyaml",
        "tabulate",
        "toml",
        "tqdm",
        "yachalk",
        "yarl",
	],
	entry_points = {
        'console_scripts': ['mangdl=mangdl.cli:cli'],
    },
)
